#!/usr/bin/python
########################################################################
#
# File:   fbcore.py
# Date:   24.05.2005
#
# Contents:
#   Basic Firebird classes.
#
# Contributors: Pavel Cisar
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# conECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
########################################################################

########################################################################
# imports
########################################################################

import os
import sys
import types
import cmd
import fbcore as fb
import kinterbasdb as kdb
from configobj import ConfigObj

########################################################################
#--- constants
########################################################################

FBCONSOLE_VERSION = "0.2"

SQL_FORMAT_SEPARATOR  = '%%'
SQL_PARAM_SEPARATOR   = '<<'
SQL_STORE_SEPARATOR   = '>>'

# Config Keys
DATABASES_SECTION        = 'databases'
DEFAULT_DB_SUBSECTION    = 'db'
DSN_KEY                  = 'dsn'
USER_KEY                 = 'user'
PASSWORD_KEY             = 'password'

########################################################################
# classes
########################################################################


## This class is based on console.py v1.0 written by James Thiele, 27 April 2004

class console(cmd.Cmd):

    def __init__(self,configObj):
        cmd.Cmd.__init__(self)
        self.prompt = "===>"
        self.contprompt = "...>"
        self.term = ';'
        self._mainDB = None
        self._mainDBName = ''
        self.intro  = "Firebird Power Console v%s\n\nType 'help' for command list.\n" % FBCONSOLE_VERSION

        # Load configuration
        self.config = configObj
        if not self.config.has_key(DATABASES_SECTION):
            self.config[DATABASES_SECTION] = {}

    #--- Support routines
    def _readLine(self):
        if self.cmdqueue:
            line = self.cmdqueue.pop(0)
        else:
            if self.use_rawinput:
                try:
                    line = raw_input(self.contprompt)
                except EOFError:
                    line = 'EOF'
            else:
                self.stdout.write(self.prompt)
                self.stdout.flush()
                line = self.stdin.readline()
                if not len(line):
                    line = 'EOF'
                else:
                    line = line[:-1] # chop \n
        return line
    def _readSQLLine(self):
        if self.cmdqueue:
            line = self.cmdqueue.pop(0)
        else:
            if self.use_rawinput:
                try:
                    line = raw_input(self.contprompt)
                except EOFError:
                    line = 'EOF'
            else:
                self.stdout.write(self.prompt)
                self.stdout.flush()
                line = self.stdin.readline()
                if not len(line):
                    line = 'EOF'
                else:
                    line = line[:-1] # chop \n
        return line
    def _dataPrinter(self,cur):
        # Print a header.
        for fieldDesc in cur.description:
            print fieldDesc[kdb.DESCRIPTION_NAME].ljust(fieldDesc[kdb.DESCRIPTION_DISPLAY_SIZE]) ,
        print
        for fieldDesc in cur.description:
            print "-" * max((len(fieldDesc[kdb.DESCRIPTION_NAME]),fieldDesc[kdb.DESCRIPTION_DISPLAY_SIZE])),
        print
        # For each row, print the value of each field left-justified within
        # the maximum possible width of that field.
        fieldIndices = range(len(cur.description))
        for row in cur:
            for fieldIndex in fieldIndices:
                fieldValue = str(row[fieldIndex])
                fieldMaxWidth = max((len(cur.description[fieldIndex][kdb.DESCRIPTION_NAME]),cur.description[fieldIndex][kdb.DESCRIPTION_DISPLAY_SIZE]))
                print fieldValue.ljust(fieldMaxWidth) ,
            print

    def _doIndexAction(self,index,action,args={}):
        if hasattr(index,action):
            getattr(index,action)(**args)
    def _indexAction(self,line,action,errorMsg):
        if line == '':
            line = self._mainDBName
        try:
            obj = eval(line,self._globals,self._locals)
        except Exception, e:
            print e.__class__, ":", e
            return
        if isinstance(obj,fb.FBMDatabase):
            obj = obj.indices
        elif isinstance(obj,fb.FBMTable):
            obj = obj.indices
        elif isinstance(obj,fb.FBMIndex):
            obj = (obj,)
        if (isinstance(obj,types.ListType) or isinstance(obj,types.TupleType)) and len(obj) > 0:
            for o in obj:
                if isinstance(o,types.StringType):
                    name = o
                    o = self._mainDB.index(name)
                    if o is None:
                        print >> sys.stderr, "Index %s not found in database %s" % (name,self._mainDBName)
                        continue
                if isinstance(o,fb.FBMTable):
                    for i in o.indices:
                        self._doIndexAction(i,action)
                elif isinstance(o,fb.FBMIndex):
                    self._doIndexAction(o,action)
                elif isinstance(o,fb.FBMDatabase):
                    for i in o.indices:
                        self._doIndexAction(i,action)
                else:
                    print >> sys.stderr, errorMsg % type(o).__name__

    #--- Override methods in Cmd object
    def preloop(self):
        """Initialization before prompting user for commands.
           Despite the claims in the Cmd documentaion, Cmd.preloop() is not a stub.
        """
        cmd.Cmd.preloop(self)   ## sets up command completion
        self._locals  = {}      ## Initialize execution namespace for user
        self._globals = {}

    def postloop(self):
        """Take care of any unfinished business.
           Despite the claims in the Cmd documentaion, Cmd.postloop() is not a stub.
        """
        cmd.Cmd.postloop(self)   ## Clean up command completion
        print "Exiting..."

    def precmd(self, line):
        """ This method is called after the line has been input but before
            it has been interpreted. If you want to modifdy the input line
            before execution (for example, variable substitution) do it here.
        """
        if line[0] == '@':
            line = 'sql ' + line[1:]
        return line

    def postcmd(self, stop, line):
        """If you want to stop the console, return something that evaluates to true.
           If you want to do some post command processing, do it here.
        """
        return stop

    def emptyline(self):
        """Do nothing on empty input line"""
        pass

    # Python command processing

    def default(self, line):
        """Called on an input line when the command prefix is not recognized.
           In that case we execute the line as Python code.
        """
        try:
            if line[-1:] == ':':
                multiline = [line,]
                stop = False
                while not stop:
                    line = self._readLine()
                    stop = (line == 'EOF') or (line.strip() == '')
                    if not stop:
                        multiline.append(line)
                    if line == 'EOF':
                        self.cmdqueue.append(line)
                line = '\n'.join(multiline)
            exec(line) in self._globals, self._locals
        except Exception, e:
            print e.__class__, ":", e
        #exec(line) in self._globals, self._locals

    #--- Command definitions to support Cmd object functionality
    def do_exit(self, args):
        """Exits from the console"""
        return -1
    def do_EOF(self, args):
        """Exit on system end of file character"""
        return self.do_exit(args)
    def do_shell(self, args):
        """Pass command to a system shell when line begins with '!'"""
        os.system(args)
    def do_help(self, args):
        """Get help on commands
    'help' or '?' with no arguments prints a list of commands for which help
    is available.
    'help <command>' or '? <command>' gives help on <command>
        """
        ## The only reason to define this method is for the help text in the doc string
        cmd.Cmd.do_help(self, args)

    #--- FB console commands

    def do_pyhelp(self,line):
        """Standard built-in Python Help System.
    Without arguments, it enters interactive Help console, otherwise it
    will show information about specified Python object."""
        if line == "":
            help()
        else:
            help(eval(line,self._globals, self._locals))
    def do_ready(self,line):
        """Attach to the database.

    usage: ready [<alias>|<db_spec>[ AS <alias>]]

    Without argument, it would try to open default alias ("db"). if this alias
    doesn't exists, it would ask for database location, user and password,
    and would store it under default alias for latter reuse.

    If any argument is specified, then it attempts to locate an alias with
    the same name, and connect to that database defined by alias. If argument
    is not recognized as alias, it's supposed to be a database connect string
    (or alias defined at the server), and prompts for user name and password.
    If connection is sucessfuly established, database path, user name and
    password are stored under default alias "db" or as alias specified by
    optional AS argument.

    Examples:
    ===>ready  ## connects to default alias "db" or creates one
    ===>ready mydb  ## connects to alias 'mysdb' or database 'mydb'
    ===>ready c:/db/employee.fdb ## connects to database and define "db" alias
    ===>ready myserver:/home/db/mydb.fdb as mydb  ## creates alias "mydb"
        """
        # First check whether we're asked to open db from config by its name
        configName = None
        dsnValue = None
        userValue = None
        passwordValue = None
        if line == '':
            configName = 'db'
        elif self.config[DATABASES_SECTION].has_key(line):
            configName = line
        if configName is not None:
            if self.config[DATABASES_SECTION].has_key(configName):
                try:
                    dsnValue = self.config[DATABASES_SECTION][configName][DSN_KEY]
                    userValue = self.config[DATABASES_SECTION][configName][USER_KEY]
                    passwordValue = self.config[DATABASES_SECTION][configName][PASSWORD_KEY]
                except:
                    pass
                name = configName
            else:
                raise Exception('Database alias %s not defined in config' % configName)
        else:
            dsnValue = None
            if line.rfind(' as ') >= 0:
                dsnValue, name = line.split(' as ')
            else:
                dsnValue = line
                name = 'db'
            if dsnValue is None:
                dsnValue = raw_input('Path to database: ')
            userValue = raw_input('User: ')
            passwordValue = raw_input('Password: ')
        d = fb.FBMDatabase()
        try:
            d.attachDatabase(dsn=dsnValue,user=userValue,password=passwordValue)
            self._locals[name] = d
            self._mainDB = d
            self._mainDBName = name
            self.config[DATABASES_SECTION][name] = {DSN_KEY: dsnValue, USER_KEY: userValue, PASSWORD_KEY: passwordValue}
            self.config.write()
            print 'Database %s at %s, user %s' % (name, dsnValue, userValue)
##            for k,v in d.getLocals().iteritems():
##                self._locals[k] = v
        except Exception, e:
            print e.__class__, ":", e
    def do_list(self,line):
        """Display name(s) of database object(s).

    usage: list <db_object>|<list_of_db_objects>

    Examples:
    ===>list db.tables  ## List names of all user tables in database
        """
        try:
            obj = eval(line,self._globals,self._locals)
        except Exception, e:
            print e.__class__, ":", e
            return
        if isinstance(obj,fb.FBMBaseDBObject):
            obj = (obj,)
        if (isinstance(obj,types.ListType) or isinstance(obj,types.TupleType)) and len(obj) > 0:
            for o in obj:
                if hasattr(o,'name'): print o.name

    def do_sql(self,line):
        """Executes SQL command.
        """
        try:
            if line[-len(self.term):] != self.term:
                multiline = [line,]
                stop = False
                while not stop:
                    line = self._readLine()
                    stop = (line == 'EOF') or (line[-len(self.term):] == self.term)
                    if line == 'EOF':
                        self.cmdqueue.append(line)
                    else:
                        multiline.append(line)
                line = '\n'.join(multiline)
            line = line[:-len(self.term)]
            c = self._mainDB.cursor()
            # check for store variable
            if line.find(SQL_STORE_SEPARATOR) >= 0:
                line,var = line.split(SQL_STORE_SEPARATOR)
                var = var.strip()
            else:
                var = None
            # Check for input parameters
            if line.find(SQL_FORMAT_SEPARATOR) >= 0:
                sql,params = line.split(SQL_FORMAT_SEPARATOR)
                values = eval(params,self._globals,self._locals)
                if type(values) is not types.TupleType:
                    values = tuple(values)
                c.execute(sql % values)
            elif line.find(SQL_PARAM_SEPARATOR) >= 0:
                sql,params = line.split(SQL_PARAM_SEPARATOR)
                params = eval(params,self._globals,self._locals)
                if (type(params) is types.GeneratorType) or isinstance(params,kdb.Cursor):
                    for par in params:
                        c.execute(sql,par)
                elif type(params[0]) is types.ListType:
                    c.executemany(sql,params)
                else:
                    c.execute(sql,params)
            else:
                c.execute(line)
            if var is not None:
                self._locals[var] = c
            elif line.upper().startswith('SELECT'):
                self._dataPrinter(c)
        except Exception, e:
            print e.__class__, ":", e

    def do_set(self,line):
        """Show or set environment options.

        usage: set  ## Display current environment settings
               set db <open-alias>  ## Set main working database
        """
        if line == '':
            for name, i in ((name, var) for name, var in self._locals.items() if isinstance(var,fb.FBMDatabase)):
                print 'Database %s at %s, user %s' % (name, i.dsn, i.user)
            print "\nCurrent database: %s" % self._mainDBName
        else:
            cmd = line.split(' ')
            if cmd[0] == 'db':
                if len(cmd) != 2:
                    print >> sys.stderr, "'set db' takes one argument (%i given)" % (len(cmd)-1)
                else:
                    db = line.split(' ')[1]
                    if self._locals.has_key(db):
                        self._mainDBName = db
                        self._mainDB = self._locals[db]
                    else:
                        print >> sys.stderr, "Database '%s' isn't ready" % db

    #--- Specific commands

    def do_commit(self,line):
        """Commits transaction.
        """
        if line == '':
            self._mainDB.commit()
        else:
            self._locals[line].commit()

    def do_rollback(self,line):
        """Rollback transaction.
        """
        if line == '':
            self._mainDB.rollback()
        else:
            self._locals[line].rollback()

    def do_recompute(self,line):
        """Recomputes index statistics.

    usage: recompute [<database>|<list_of_indices>|<list_of_names>]

    Without argument it would recompute all user indices in working database.
    If argument is a database object, it would recompute all user indices in it
    If argument is a table object, it would recompute indices for this table.
    If argument is a list of index or table objects, it would recompute these
    indices or indices associated to these tables.
    If argument is a list object, it must contain names of indices in current
    database to recompute.

    Examples:
    ===>recompute employee  ## recompute all in employee database
    ===>recompute db.indices  ## recompute all indices in database db
    ===>recompute db.table('JOB').indices ## recompute indices of JOB table
    ===>recompute ['JOB_FK1','JOB_FK2']  ## recomputes specified indices
        """
        self._indexAction(line,'recompute',"Can't recompute statistics for '%s' object")

    def do_deactivate(self,line):
        """Deactivates indices.

    usage: deactivate [<database_obj>|<table_obj>|<index_obj>|<list>]

    Without argument it would deactivate all user indices in working database.
    If argument is a database object, it would deactivate all user indices in it
    If argument is a table object, it would deactivate indices for this table.
    If argument is list of index or table objects, it would deactivate these
    indices or indices associated to these tables.
    If argument is list object, it must contain names of indices in current
    database to deactivate.

    Examples:
    ===>deactivate employee  ## deactivates all in employee database
    ===>deactivate db.indices  ## deactivates all indices in database db
    ===>deactivate db.table('JOB').indices ## deactivates indices of JOB table
    ===>deactivate ['JOB_FK1','JOB_FK2']  ## deactivates specified indices
        """
        self._indexAction(line,'deactivate',"Can't deactivate indices for '%s' object")

    def do_activate(self,line):
        """Deactivates indices.

    usage: activate [<database_obj>|<table_obj>|<index_obj>|<list>]

    Without argument it would activate all user indices in working database.
    If argument is a database object, it would activate all user indices in it
    If argument is a table object, it would activate indices for this table.
    If argument is list of index or table objects, it would activate these
    indices or indices associated to these tables.
    If argument is list object, it must contain names of indices in current
    database to activate.

    Examples:
    ===>activate employee  ## activates all in employee database
    ===>activate db.indices  ## activates all indices in database db
    ===>activate db.table('JOB').indices ## activates indices of JOB table
    ===>activate ['JOB_FK1','JOB_FK2']  ## activates specified indices
        """
        self._indexAction(line,'activate',"Can't activate indices for '%s' object")

    def do_rebuild(self,line):
        """Deactivates indices.

    usage: rebuild [<database_obj>|<table_obj>|<index_obj>|<list>]

    Without argument it would rebuild all user indices in working database.
    If argument is a database object, it would rebuild all user indices in it
    If argument is a table object, it would rebuild indices for this table.
    If argument is list of index or table objects, it would rebuild these
    indices or indices associated to these tables.
    If argument is list object, it must contain names of indices in current
    database to rebuild.

    Examples:
    ===>rebuild employee  ## rebuilds all in employee database
    ===>rebuild db.indices  ## rebuilds all indices in database db
    ===>rebuild db.table('JOB').indices ## rebuilds indices of JOB table
    ===>rebuild ['JOB_FK1','JOB_FK2']  ## rebuilds specified indices
        """
        self._indexAction(line,'rebuild',"Can't rebuild indices for '%s' object")

#--- Main program

if __name__ == '__main__':
        config = ConfigObj(os.path.expanduser('~/console.cfg'),{'create_empty':True})
        console = console(config)
        console.cmdloop()
        config.write()
